@@ -1,6 +1,6 @@
 use criterion::{criterion_group, criterion_main, Criterion};
 use hgrep::chunk::{File, LineMatch};
-use hgrep::printer::{Printer, PrinterOptions, TermColorSupport};
+use hgrep::printer::{Printer, PrinterOptions, TermColorSupport, TextWrapMode};
 use hgrep::ripgrep;
 use hgrep::syntect::{LockableWrite, SyntectPrinter};
 use hgrep_bench::node_modules_path;
@@ -33,6 +33,13 @@ impl<'a> LockableWrite<'a> for Sink {
     }
 }
 
+fn get_opts() -> PrinterOptions<'static> {
+    let mut opts = PrinterOptions::default();
+    opts.color_support = TermColorSupport::True;
+    opts.term_width = 80;
+    opts
+}
+
 fn large_file(c: &mut Criterion) {
     let (path, contents) = read_package_lock_json();
     let lines = contents.lines().count() as u64;
@@ -50,10 +57,8 @@ fn large_file(c: &mut Criterion) {
 
     c.bench_function("syntect::package-lock.json", |b| {
         b.iter(|| {
-            let mut opts = PrinterOptions::default();
-            opts.color_support = TermColorSupport::True;
-            opts.term_width = 80;
             let sink = Sink(Mutex::new(vec![]));
+            let opts = get_opts();
             let mut printer = SyntectPrinter::new(sink, opts).unwrap();
             files
                 .clone()
@@ -68,10 +73,7 @@ fn large_file(c: &mut Criterion) {
 
 fn with_ripgrep(c: &mut Criterion) {
     #[inline]
-    fn run_ripgrep(pat: &str, dir: &Path) -> bool {
-        let mut opts = PrinterOptions::default();
-        opts.color_support = TermColorSupport::True;
-        opts.term_width = 80;
+    fn run_ripgrep(pat: &str, dir: &Path, opts: PrinterOptions<'_>) -> bool {
         let sink = Sink(Mutex::new(vec![]));
         let printer = SyntectPrinter::new(sink, opts).unwrap();
         let mut config = ripgrep::Config::new(3, 6);
@@ -79,19 +81,35 @@ fn with_ripgrep(c: &mut Criterion) {
         ripgrep::grep(printer, pat, Some(iter::once(dir.as_os_str())), config).unwrap()
     }
 
-    let dir = node_modules_path();
+    let node_modules = node_modules_path();
     c.bench_function("syntect::ripgrep-large", |b| {
-        b.iter(|| assert!(run_ripgrep(r"\bparcel\b", dir)))
+        b.iter(|| assert!(run_ripgrep(r"\bparcel\b", node_modules, get_opts())))
     });
 
-    let dir = Path::new("..").join("src");
+    let project_src = Path::new("..").join("src");
     c.bench_function("syntect::ripgrep-small", |b| {
-        b.iter(|| assert!(run_ripgrep("Printer", &dir)))
+        b.iter(|| assert!(run_ripgrep("Printer", &project_src, get_opts())))
     });
 
-    let dir = Path::new("..").join("testdata").join("chunk");
+    let testdata = Path::new("..").join("testdata").join("chunk");
     c.bench_function("syntect::ripgrep-tiny", |b| {
-        b.iter(|| assert!(run_ripgrep(r"\*$", &dir)))
+        b.iter(|| assert!(run_ripgrep(r"\*$", &testdata, get_opts())))
+    });
+
+    c.bench_function("syntect::ripgrep-no-wrap", |b| {
+        b.iter(|| {
+            let mut opts = get_opts();
+            opts.text_wrap = TextWrapMode::Never;
+            assert!(run_ripgrep("Printer", &project_src, opts))
+        })
+    });
+
+    c.bench_function("syntect::ripgrep-background", |b| {
+        b.iter(|| {
+            let mut opts = get_opts();
+            opts.background_color = true;
+            assert!(run_ripgrep("Printer", &project_src, opts))
+        })
     });
 }
 
