@@ -7,6 +7,9 @@ use std::env;
 use std::io;
 use std::process;
 
+#[global_allocator]
+static GLOBAL: mimalloc::MiMalloc = mimalloc::MiMalloc;
+
 #[cfg(feature = "ripgrep")]
 use hgrep::ripgrep;
 
@@ -107,6 +110,11 @@ fn cli<'a>() -> App<'a> {
                 .possible_values(["char", "never"])
                 .case_insensitive(true)
                 .about("Text-wrapping mode. 'char' enables character-wise text-wrapping. 'never' disables text-wrapping")
+        ).arg(
+            Arg::new("first-only")
+                .short('f')
+                .long("first-only")
+                .about("Show only the first code snippet per file")
         )
         .arg(
             Arg::new("generate-completion-script")
@@ -126,11 +134,17 @@ fn cli<'a>() -> App<'a> {
     );
 
     #[cfg(feature = "syntect-printer")]
-    let app = app.arg(
-        Arg::new("background")
-            .long("background")
-            .about("Paint background colors. This flag is only for syntect printer"),
-    );
+    let app = app
+        .arg(
+            Arg::new("background")
+                .long("background")
+                .about("Paint background colors. This flag is only for syntect printer"),
+        )
+        .arg(
+            Arg::new("ascii-lines").long("ascii-lines").about(
+                "Use ascii characters for drawing border lines instead of Unicode characters",
+            ),
+        );
 
     #[cfg(feature = "ripgrep")]
     let app = app
@@ -419,12 +433,26 @@ fn app() -> Result<bool> {
         printer_opts.text_wrap = TextWrapMode::Never;
     }
 
+    if matches.is_present("first-only") {
+        printer_opts.first_only = true;
+    }
+
     #[cfg(feature = "syntect-printer")]
-    if matches.is_present("background") {
-        printer_opts.background_color = true;
-        #[cfg(feature = "bat-printer")]
-        if printer_kind == PrinterKind::Bat {
-            anyhow::bail!("--background flag is only available for syntect printer since bat does not support painting background colors");
+    {
+        if matches.is_present("background") {
+            printer_opts.background_color = true;
+            #[cfg(feature = "bat-printer")]
+            if printer_kind == PrinterKind::Bat {
+                anyhow::bail!("--background flag is only available for syntect printer since bat does not support painting background colors");
+            }
+        }
+
+        if matches.is_present("ascii-lines") {
+            printer_opts.ascii_lines = true;
+            #[cfg(feature = "bat-printer")]
+            if printer_kind == PrinterKind::Bat {
+                anyhow::bail!("--ascii-lines flag is only available for syntect printer since bat does not support this feature");
+            }
         }
     }
 
@@ -561,8 +589,7 @@ fn main() {
         Ok(true) => 0,
         Ok(false) => 1,
         Err(err) => {
-            let red = ansi_term::Colour::Red.bold();
-            eprintln!("{} {}", red.paint("error:"), err);
+            eprintln!("\x1b[1;91merror:\x1b[0m {}", err);
             2
         }
     };
