@@ -2,6 +2,7 @@ use crate::chunk::File;
 use crate::chunk::Line;
 use crate::printer::{Printer, PrinterOptions, TermColorSupport, TextWrapMode};
 use anyhow::Result;
+use flate2::read::ZlibDecoder;
 use memchr::{memchr_iter, Memchr};
 use rgb2ansi256::rgb_to_ansi256;
 use std::cmp;
@@ -12,6 +13,7 @@ use std::io::Write;
 use std::io::{self, Stdout, StdoutLock};
 use std::ops::{Deref, DerefMut};
 use std::path::Path;
+use std::str::Chars;
 use syntect::highlighting::{
     Color, FontStyle, HighlightIterator, HighlightState, Highlighter, Style, Theme, ThemeSet,
 };
@@ -39,7 +41,7 @@ impl<'a> LockableWrite<'a> for Stdout {
 
 pub fn list_themes<W: Write>(mut out: W) -> Result<()> {
     let mut seen = HashSet::new();
-    let bat_defaults = bincode::deserialize_from(flate2::read::ZlibDecoder::new(THEME_SET_BIN))?;
+    let bat_defaults = bincode::deserialize_from(ZlibDecoder::new(THEME_SET_BIN))?;
     let defaults = ThemeSet::load_defaults();
     for themes in &[bat_defaults, defaults] {
         for name in themes.themes.keys() {
@@ -88,74 +90,122 @@ struct Token<'line> {
     text: &'line str,
 }
 
-// Start/End boundary of the match region in the highlighted text in matched line
-struct RegionBoundary {
-    start: Option<usize>, // This index is local to the text
-    end: Option<usize>,   // This index is local to the text
-    // Original style of the text. Background color is not necessary because it is fixed to match color
-    fg: Color,
-}
-
-// Match region in matched line
-struct Region(usize, usize);
-
-impl Region {
-    fn is_empty(&self) -> bool {
-        self.0 == self.1
-    }
-
-    fn sub(&self, diff: usize) -> Self {
-        Self(self.0.saturating_sub(diff), self.1.saturating_sub(diff))
-    }
-
-    // Calculate byte indices of start/end of the region within this text if they are included
-    fn boundary(&self, token_start: usize, token_end: usize, fg: Color) -> Option<RegionBoundary> {
-        let (rs, re) = (self.0, self.1);
-        let start = (token_start <= rs && rs <= token_end).then(|| rs - token_start);
-        let end = (token_start <= re && re <= token_end).then(|| re - token_start);
-        if start.is_none() && end.is_none() {
-            return None;
-        }
-        Some(RegionBoundary { start, end, fg })
-    }
-
-    fn contains(&self, byte_offset: usize) -> bool {
-        self.0 <= byte_offset && byte_offset <= self.1
-    }
-}
-
-enum Matched {
-    Region(Region),
-    Yes,
-    No,
-}
-impl Matched {
-    fn yes(&self) -> bool {
-        match self {
-            Matched::Region(..) => true,
-            Matched::Yes => true,
-            Matched::No => false,
+impl<'line> Token<'line> {
+    fn chomp(&mut self) {
+        if self.text.ends_with('\n') {
+            self.text = &self.text[..self.text.len() - 1];
+            if self.text.ends_with('\r') {
+                self.text = &self.text[..self.text.len() - 1];
+            }
         }
     }
-    fn region(&self) -> Option<Option<&Region>> {
-        match self {
-            Matched::Region(r) => Some(Some(r)),
-            Matched::Yes => Some(None),
-            Matched::No => None,
+}
+
+#[derive(Clone, Copy)]
+enum RegionBoundary {
+    Start,
+    End,
+    NotFound,
+}
+
+enum DrawEvent {
+    RegionStart,
+    RegionEnd,
+    Char(char),
+    TokenBoundary(Style), // Previous style
+    Done,
+}
+
+struct DrawEvents<'a, 'line: 'a> {
+    tokens: &'a [Token<'line>],
+    chars_in_token: Chars<'line>,
+    regions: &'a [(usize, usize)],
+    current_style: Style,
+    in_region: bool,
+    byte_offset: usize,
+}
+
+impl<'a, 'line: 'a> DrawEvents<'a, 'line> {
+    fn new(tokens: &'a [Token<'line>], regions: &'a [(usize, usize)]) -> Self {
+        let (chars_in_token, current_style, tokens) =
+            if let Some((head, tail)) = tokens.split_first() {
+                (head.text.chars(), head.style, tail)
+            } else {
+                ("".chars(), Style::default(), tokens)
+            };
+
+        Self {
+            tokens,
+            chars_in_token,
+            regions,
+            current_style,
+            in_region: false,
+            byte_offset: 0,
+        }
+    }
+
+    fn region_boundary(&mut self) -> RegionBoundary {
+        let o = self.byte_offset;
+
+        // Eat done regions
+        let num_done_regions = self.regions.iter().take_while(|(_, e)| *e < o).count();
+        if num_done_regions > 0 {
+            self.regions = &self.regions[num_done_regions..];
+        }
+
+        match self.regions.first().copied() {
+            Some((s, e)) if o == s && o < e => RegionBoundary::Start,
+            Some((_, e)) if o == e => {
+                // When the next region is adjcent, skip changing highlight
+                match self.regions.get(1) {
+                    Some((s, _)) if o == *s => RegionBoundary::NotFound,
+                    _ => RegionBoundary::End,
+                }
+            }
+            _ => RegionBoundary::NotFound,
+        }
+    }
+
+    fn next_event(&mut self) -> DrawEvent {
+        match self.region_boundary() {
+            RegionBoundary::Start if !self.in_region => {
+                self.in_region = true;
+                return DrawEvent::RegionStart;
+            }
+            RegionBoundary::End if self.in_region => {
+                self.in_region = false;
+                return DrawEvent::RegionEnd;
+            }
+            _ => { /* fall through */ }
+        }
+
+        if let Some(c) = self.chars_in_token.next() {
+            self.byte_offset += c.len_utf8();
+            return DrawEvent::Char(c);
+        }
+
+        if let Some((head, tail)) = self.tokens.split_first() {
+            let prev_style = self.current_style;
+            self.current_style = head.style;
+            self.chars_in_token = head.text.chars();
+            self.tokens = tail;
+            DrawEvent::TokenBoundary(prev_style)
+        } else {
+            DrawEvent::Done
         }
     }
 }
 
 struct Canvas<'file, W: Write> {
     out: W,
-    tab_width: u16,
     theme: &'file Theme,
     true_color: bool,
-    background: bool,
-    match_color: Option<Color>,
-    region_fg_color: Option<Color>,
-    region_bg_color: Option<Color>,
-    wrap: bool,
+    has_background: bool,
+    match_bg: Option<Color>,
+    region_fg: Option<Color>,
+    region_bg: Option<Color>,
+    current_fg: Option<Color>,
+    current_bg: Option<Color>,
 }
 
 impl<'file, W: Write> Deref for Canvas<'file, W> {
@@ -170,15 +220,6 @@ impl<'file, W: Write> DerefMut for Canvas<'file, W> {
     }
 }
 
-enum LineDrawState<'line> {
-    Continue(usize),
-    Break(&'line str),
-}
-enum LineDrawn<'line> {
-    Done,
-    Wrap(usize, &'line str, usize),
-}
-
 impl<'file, W: Write> Canvas<'file, W> {
     fn draw_spaces(&mut self, num: usize) -> Result<()> {
         for _ in 0..num {
@@ -189,6 +230,8 @@ impl<'file, W: Write> Canvas<'file, W> {
 
     fn draw_newline(&mut self) -> Result<()> {
         writeln!(self.out, "\x1b[0m")?; // Reset on newline to ensure to reset color
+        self.current_fg = None;
+        self.current_bg = None;
         Ok(())
     }
 
@@ -213,15 +256,23 @@ impl<'file, W: Write> Canvas<'file, W> {
     }
 
     fn set_bg(&mut self, c: Color) -> Result<()> {
-        self.set_color(40, c)
+        if self.current_bg != Some(c) {
+            self.set_color(40, c)?;
+            self.current_bg = Some(c);
+        }
+        Ok(())
     }
 
     fn set_fg(&mut self, c: Color) -> Result<()> {
-        self.set_color(30, c)
+        if self.current_fg != Some(c) {
+            self.set_color(30, c)?;
+            self.current_fg = Some(c);
+        }
+        Ok(())
     }
 
     fn set_default_bg(&mut self) -> Result<()> {
-        if self.background {
+        if self.has_background {
             if let Some(bg) = self.theme.settings.background {
                 self.set_bg(bg)?;
             }
@@ -229,6 +280,20 @@ impl<'file, W: Write> Canvas<'file, W> {
         Ok(())
     }
 
+    fn set_default_fg(&mut self) -> Result<()> {
+        if let Some(fg) = self.theme.settings.foreground {
+            self.set_fg(fg)?;
+        }
+        Ok(())
+    }
+
+    fn set_background(&mut self, c: Color) -> Result<()> {
+        if self.has_background {
+            self.set_bg(c)?;
+        }
+        Ok(())
+    }
+
     fn set_bold(&mut self) -> Result<()> {
         self.out.write_all(b"\x1b[1m")?;
         Ok(())
@@ -259,220 +324,71 @@ impl<'file, W: Write> Canvas<'file, W> {
         Ok(())
     }
 
+    fn set_style(&mut self, style: Style) -> Result<()> {
+        self.set_background(style.background)?;
+        self.set_fg(style.foreground)?;
+        self.set_font_style(style.font_style)?;
+        Ok(())
+    }
+
     fn set_match_bg_color(&mut self) -> Result<()> {
-        if let Some(bg) = self.match_color {
+        if let Some(bg) = self.match_bg {
             self.set_bg(bg)?;
         }
         Ok(())
     }
 
+    fn set_match_style(&mut self, style: Style) -> Result<()> {
+        self.set_fg(style.foreground)?;
+        self.set_match_bg_color()?;
+        self.set_font_style(style.font_style)
+    }
+
     fn set_region_color(&mut self) -> Result<()> {
-        if let Some(c) = self.region_fg_color {
+        if let Some(c) = self.region_fg {
             self.set_fg(c)?;
         }
-        if let Some(c) = self.region_bg_color {
+        if let Some(c) = self.region_bg {
             self.set_bg(c)?;
         }
         Ok(())
     }
 
-    fn set_boundary_color(&mut self, boundary: &RegionBoundary, offset: usize) -> Result<()> {
-        if boundary.start == Some(offset) {
-            self.set_region_color()?;
-        }
-        if boundary.end == Some(offset) {
-            self.set_fg(boundary.fg)?;
-            self.set_match_bg_color()?;
-        }
-        Ok(())
-    }
-
-    // Returns number of tab characters in the text
-    fn draw_text<'line>(
-        &mut self,
-        text: &'line str,
-        limit: usize,
-        boundary: Option<RegionBoundary>,
-    ) -> Result<LineDrawState<'line>> {
-        let mut width = 0;
-        let mut saw_zwj = false;
-        for (i, c) in text.char_indices() {
-            if let Some(boundary) = &boundary {
-                self.set_boundary_color(boundary, i)?;
-            }
-
-            width += if c == '\t' && self.tab_width > 0 {
-                let w = self.tab_width as usize;
-                if width + w > limit {
-                    self.draw_spaces(limit - width)?;
-                    // `+ 1` for skipping rest of \t
-                    return Ok(LineDrawState::Break(&text[i + 1..]));
-                }
-                self.draw_spaces(self.tab_width as usize)?;
-                w
-            } else {
-                // Handle zero width joiner
-                let w = if c == '\u{200d}' {
-                    saw_zwj = true;
-                    0
-                } else if saw_zwj {
-                    saw_zwj = false;
-                    0 // Do not count width while joining current character into previous one with ZWJ
-                } else {
-                    c.width_cjk().unwrap_or(0)
-                };
-                if width + w > limit {
-                    self.draw_spaces(limit - width)?;
-                    return Ok(LineDrawState::Break(&text[i..]));
-                }
-                write!(self.out, "{}", c)?;
-                w
-            };
-        }
-        Ok(LineDrawState::Continue(width))
-    }
-
-    fn draw_text_no_wrap(&mut self, text: &str) -> Result<usize> {
-        if self.tab_width == 0 {
-            write!(self.out, "{}", text)?;
-            return Ok(text.width_cjk());
-        }
-
-        let mut width = 0;
-        for c in text.chars() {
-            if c == '\t' && self.tab_width > 0 {
-                let w = self.tab_width as usize;
-                self.draw_spaces(w)?;
-                width += w;
-            } else {
-                write!(self.out, "{}", c)?;
-                width += c.width_cjk().unwrap_or(0);
-            }
-        }
-        Ok(width)
-    }
-
-    fn draw_text_no_wrap_with_region(
-        &mut self,
-        text: &str,
-        boundary: RegionBoundary,
-    ) -> Result<usize> {
-        let mut width = 0;
-        for (i, c) in text.chars().enumerate() {
-            self.set_boundary_color(&boundary, i)?;
-            if c == '\t' && self.tab_width > 0 {
-                let w = self.tab_width as usize;
-                self.draw_spaces(w)?;
-                width += w;
-            } else {
-                write!(self.out, "{}", c)?;
-                width += c.width_cjk().unwrap_or(0);
-            }
-        }
-        Ok(width)
-    }
-
     fn fill_spaces(&mut self, written_width: usize, max_width: usize) -> Result<()> {
         if written_width < max_width {
             self.draw_spaces(max_width - written_width)?;
         }
         Ok(())
     }
-
-    fn draw_matched<'line>(
-        &mut self,
-        region: Option<&Region>,
-        tokens: &[Token<'line>],
-        max_width: usize,
-    ) -> Result<LineDrawn<'line>> {
-        self.set_match_bg_color()?;
-
-        let mut start_offset = 0;
-        let mut width = 0;
-        for (idx, tok) in tokens.iter().enumerate() {
-            let len = tok.text.len();
-            let end_offset = start_offset + len;
-
-            let boundary = if let Some(region) = region {
-                // In region, the style should not be changed
-                if !region.contains(start_offset) {
-                    self.set_fg(tok.style.foreground)?;
-                    self.set_font_style(tok.style.font_style)?;
-                }
-                region.boundary(start_offset, end_offset, tok.style.foreground)
-            } else {
-                self.set_fg(tok.style.foreground)?;
-                None
-            };
-
-            if self.wrap {
-                match self.draw_text(tok.text, max_width - width, boundary)? {
-                    LineDrawState::Continue(w) => width += w,
-                    LineDrawState::Break(rest) => {
-                        let bytes = end_offset - rest.len();
-                        return Ok(LineDrawn::Wrap(bytes, rest, idx));
-                    }
-                }
-            } else if let Some(boundary) = boundary {
-                width += self.draw_text_no_wrap_with_region(tok.text, boundary)?;
-            } else {
-                width += self.draw_text_no_wrap(tok.text)?;
-            }
-            self.unset_font_style(tok.style.font_style)?;
-            start_offset += len;
-        }
-
-        // Ensure to reset background color considering the case where a region is at the end of line
-        self.set_match_bg_color()?;
-        self.fill_spaces(width, max_width)?;
-
-        Ok(LineDrawn::Done)
-    }
-
-    fn draw<'line>(
-        &mut self,
-        tokens: &[Token<'line>],
-        matched: &Matched,
-        max_width: usize,
-    ) -> Result<LineDrawn<'line>> {
-        if let Some(maybe_region) = matched.region() {
-            return self.draw_matched(maybe_region, tokens, max_width);
-        }
-
-        let mut byte_offset = 0;
-        let mut width = 0;
-        for (idx, tok) in tokens.iter().enumerate() {
-            if self.background {
-                self.set_bg(tok.style.background)?;
-            }
-            self.set_fg(tok.style.foreground)?;
-            self.set_font_style(tok.style.font_style)?;
-            if self.wrap {
-                match self.draw_text(tok.text, max_width - width, None)? {
-                    LineDrawState::Continue(w) => width += w,
-                    LineDrawState::Break(rest) => {
-                        let bytes = byte_offset + tok.text.len() - rest.len();
-                        return Ok(LineDrawn::Wrap(bytes, rest, idx));
-                    }
-                }
-            } else {
-                width += self.draw_text_no_wrap(tok.text)?;
-            }
-            self.unset_font_style(tok.style.font_style)?;
-            byte_offset += tok.text.len();
-        }
-
-        if width == 0 {
-            self.set_default_bg()?; // For empty line
-        }
-        if self.background {
-            self.fill_spaces(width, max_width)?;
-        }
-
-        Ok(LineDrawn::Done)
-    }
 }
 
+struct LineChars<'a> {
+    horizontal: &'a str,
+    vertical: &'a str,
+    vertical_and_right: &'a str,
+    down_and_horizontal: &'a str,
+    up_and_horizontal: &'a str,
+    dashed_horizontal: &'a str,
+}
+
+const UNICODE_LINE_CHARS: LineChars<'static> = LineChars {
+    horizontal: "─",
+    vertical: "│",
+    vertical_and_right: "├",
+    down_and_horizontal: "┬",
+    up_and_horizontal: "┴",
+    dashed_horizontal: "╶",
+};
+
+const ASCII_LINE_CHARS: LineChars<'static> = LineChars {
+    horizontal: "-",
+    vertical: "|",
+    vertical_and_right: "|",
+    down_and_horizontal: "-",
+    up_and_horizontal: "-",
+    dashed_horizontal: "-",
+};
+
 // Note: More flexible version of syntect::easy::HighlightLines for our use case
 struct LineHighlighter<'a> {
     hl: Highlighter<'a>,
@@ -506,6 +422,7 @@ impl<'a> LineHighlighter<'a> {
             .collect()
     }
 }
+
 // Like chunk::Lines, but includes newlines
 struct LinesInclusive<'a> {
     lnum: usize,
@@ -545,12 +462,15 @@ impl<'a> Iterator for LinesInclusive<'a> {
 
 // Drawer is responsible for one-time screen drawing
 struct Drawer<'file, W: Write> {
-    theme: &'file Theme,
     grid: bool,
     term_width: u16,
     lnum_width: u16,
     background: bool,
+    first_only: bool,
     gutter_color: Color,
+    wrap: bool,
+    tab_width: u16,
+    chars: LineChars<'file>,
     canvas: Canvas<'file, W>,
 }
 
@@ -562,14 +482,18 @@ impl<'file, W: Write> Drawer<'file, W> {
             lnum_width = cmp::max(lnum_width, 3); // Consider '...' in gutter
         }
 
-        let gutter_color = theme.settings.gutter_foreground.unwrap_or(Color {
-            r: 128,
-            g: 128,
-            b: 128,
-            a: 255,
-        });
+        let gutter_color = theme
+            .settings
+            .gutter_foreground
+            .or(theme.settings.minimap_border)
+            .unwrap_or(Color {
+                r: 128,
+                g: 128,
+                b: 128,
+                a: 255,
+            });
 
-        let (region_fg_color, region_bg_color) = if let Some(bg) = theme.settings.find_highlight {
+        let (region_fg, region_bg) = if let Some(bg) = theme.settings.find_highlight {
             (theme.settings.find_highlight_foreground, Some(bg))
         } else {
             (None, theme.settings.selection)
@@ -578,22 +502,31 @@ impl<'file, W: Write> Drawer<'file, W> {
         let canvas = Canvas {
             theme,
             true_color: opts.color_support == TermColorSupport::True,
-            tab_width: opts.tab_width as u16,
-            background: opts.background_color,
-            wrap: opts.text_wrap == TextWrapMode::Char,
-            region_fg_color,
-            region_bg_color,
-            match_color: theme.settings.line_highlight.or(theme.settings.background),
+            has_background: opts.background_color,
+            region_fg,
+            region_bg,
+            current_fg: None,
+            current_bg: None,
+            match_bg: theme.settings.line_highlight.or(theme.settings.background),
             out,
         };
 
+        let chars = if opts.ascii_lines {
+            ASCII_LINE_CHARS
+        } else {
+            UNICODE_LINE_CHARS
+        };
+
         Drawer {
-            theme,
             grid: opts.grid,
             term_width: opts.term_width,
             lnum_width,
             background: opts.background_color,
             gutter_color,
+            wrap: opts.text_wrap == TextWrapMode::Char,
+            tab_width: opts.tab_width as u16,
+            first_only: opts.first_only,
+            chars,
             canvas,
         }
     }
@@ -612,22 +545,21 @@ impl<'file, W: Write> Drawer<'file, W> {
         self.canvas.set_default_bg()?;
         let gutter_width = self.gutter_width();
         for _ in 0..gutter_width - 2 {
-            self.canvas.write_all("─".as_bytes())?;
+            self.canvas.write_all(self.chars.horizontal.as_bytes())?;
         }
         self.canvas.write_all(sep.as_bytes())?;
         for _ in 0..self.term_width - gutter_width + 1 {
-            self.canvas.write_all("─".as_bytes())?;
+            self.canvas.write_all(self.chars.horizontal.as_bytes())?;
         }
         self.canvas.draw_newline()
     }
 
     fn draw_line_number(&mut self, lnum: u64, matched: bool) -> Result<()> {
-        let fg = if matched {
-            self.theme.settings.foreground.unwrap()
+        if matched {
+            self.canvas.set_default_fg()?;
         } else {
-            self.gutter_color
-        };
-        self.canvas.set_fg(fg)?;
+            self.canvas.set_fg(self.gutter_color)?;
+        }
         self.canvas.set_default_bg()?;
         let width = num_digits(lnum);
         self.canvas
@@ -637,7 +569,7 @@ impl<'file, W: Write> Drawer<'file, W> {
             if matched {
                 self.canvas.set_fg(self.gutter_color)?;
             }
-            self.canvas.write_all(" │".as_bytes())?;
+            write!(self.canvas, " {}", self.chars.vertical)?;
         }
         self.canvas.set_default_bg()?;
         write!(self.canvas, " ")?;
@@ -649,7 +581,7 @@ impl<'file, W: Write> Drawer<'file, W> {
         self.canvas.set_default_bg()?;
         self.canvas.draw_spaces(self.lnum_width as usize + 2)?;
         if self.grid {
-            self.canvas.write_all("│ ".as_bytes())?;
+            write!(self.canvas, "{} ", self.chars.vertical)?;
         }
         Ok(())
     }
@@ -661,7 +593,7 @@ impl<'file, W: Write> Drawer<'file, W> {
         let left_margin = self.lnum_width + 1 - 3;
         self.canvas.draw_spaces(left_margin as usize)?;
         let w = if self.grid {
-            write!(self.canvas, "... ├")?;
+            write!(self.canvas, "... {}", self.chars.vertical_and_right)?;
             5
         } else {
             write!(self.canvas, "...")?;
@@ -670,52 +602,117 @@ impl<'file, W: Write> Drawer<'file, W> {
         self.canvas.set_default_bg()?;
         let body_width = self.term_width - left_margin - w; // This crashes when terminal width is smaller than gutter
         for _ in 0..body_width {
-            self.canvas.write_all("╶".as_bytes())?;
+            self.canvas
+                .write_all(self.chars.dashed_horizontal.as_bytes())?;
         }
         self.canvas.draw_newline()
     }
 
+    fn draw_text_wrappping(&mut self, matched: bool, style: Style, in_region: bool) -> Result<()> {
+        self.canvas.draw_newline()?;
+        self.draw_wrapping_gutter()?;
+        if in_region {
+            self.canvas.set_region_color()
+        } else if matched {
+            self.canvas.set_match_style(style)
+        } else {
+            self.canvas.set_style(style)
+        }
+    }
+
     fn draw_line(
         &mut self,
         mut tokens: Vec<Token<'_>>,
         lnum: u64,
-        mut matched: Matched,
+        regions: Option<Vec<(usize, usize)>>,
     ) -> Result<()> {
         // The highlighter requires newline at the end. But we don't want it since
         // - we sometimes need to fill the rest of line with spaces
         // - we clear colors before writing newline
-        if let Some(Token { text: t, .. }) = tokens.last_mut() {
-            if t.ends_with('\n') {
-                *t = &t[..t.len() - 1];
-                if t.ends_with('\r') {
-                    *t = &t[..t.len() - 1];
-                }
+        if let Some(tok) = tokens.last_mut() {
+            tok.chomp();
+            if tok.text.is_empty() {
+                tokens.pop(); // As the result of `chomp()`, text may be empty. Empty token can be removed
             }
         }
 
         let body_width = (self.term_width - self.gutter_width()) as usize;
-        self.draw_line_number(lnum, matched.yes())?;
-        let mut tokens = tokens.as_mut_slice();
+        let matched = regions.is_some();
 
-        while let LineDrawn::Wrap(bytes, rest, idx) =
-            self.canvas.draw(tokens, &matched, body_width)?
-        {
-            if let Matched::Region(region) = matched {
-                let region = region.sub(bytes);
-                matched = if region.is_empty() {
-                    Matched::Yes // Selected text was already reandered
-                } else {
-                    Matched::Region(region)
+        let tokens = tokens.as_slice();
+        let regions = regions.as_ref().map(AsRef::as_ref).unwrap_or(&[][..]);
+        let mut events = DrawEvents::new(tokens, regions);
+
+        self.draw_line_number(lnum, matched)?;
+        if matched {
+            self.canvas.set_match_style(events.current_style)?;
+        } else if !tokens.is_empty() {
+            self.canvas.set_style(events.current_style)?;
+        }
+
+        let mut width = 0; // Text width written to terminal
+        let mut saw_zwj = false;
+        loop {
+            match events.next_event() {
+                DrawEvent::Char('\t') if self.tab_width > 0 => {
+                    let w = self.tab_width as usize;
+                    if width + w > body_width && self.wrap {
+                        self.canvas.draw_spaces(body_width - width)?;
+                        self.draw_text_wrappping(matched, events.current_style, events.in_region)?;
+                        width = 0;
+                    } else {
+                        self.canvas.draw_spaces(w)?;
+                        width += w;
+                    }
                 }
+                DrawEvent::Char(c) => {
+                    // Handle zero width joiner
+                    let w = if c == '\u{200d}' {
+                        saw_zwj = true;
+                        0
+                    } else if saw_zwj {
+                        saw_zwj = false;
+                        0 // Do not count width while joining current character into previous one with ZWJ
+                    } else {
+                        c.width_cjk().unwrap_or(0)
+                    };
+                    if width + w > body_width && self.wrap {
+                        self.canvas.draw_spaces(body_width - width)?;
+                        self.draw_text_wrappping(matched, events.current_style, events.in_region)?;
+                        width = 0;
+                    }
+                    write!(self.canvas, "{}", c)?;
+                    width += w;
+                }
+                DrawEvent::TokenBoundary(prev_style) => {
+                    if !events.in_region {
+                        self.canvas.unset_font_style(prev_style.font_style)?;
+                        if !matched {
+                            self.canvas
+                                .set_background(events.current_style.background)?;
+                        }
+                        self.canvas.set_fg(events.current_style.foreground)?;
+                        self.canvas
+                            .set_font_style(events.current_style.font_style)?;
+                    }
+                }
+                DrawEvent::RegionStart => {
+                    self.canvas.set_region_color()?;
+                }
+                DrawEvent::RegionEnd => {
+                    self.canvas.set_match_style(events.current_style)?;
+                }
+                DrawEvent::Done => break,
             }
-            self.canvas.draw_newline()?;
-            self.draw_wrapping_gutter()?;
-            if rest.is_empty() {
-                tokens = &mut tokens[idx + 1..];
-            } else {
-                tokens = &mut tokens[idx..];
-                tokens[0].text = rest; // Set rest of the text broken by text wrapping
-            }
+        }
+
+        if matched {
+            self.canvas.set_match_bg_color()?;
+        } else if width == 0 {
+            self.canvas.set_default_bg()?;
+        }
+        if self.canvas.has_background || matched {
+            self.canvas.fill_spaces(width, body_width)?;
         }
 
         self.canvas.draw_newline()
@@ -737,23 +734,22 @@ impl<'file, W: Write> Drawer<'file, W> {
                 continue;
             }
             if start <= lnum && lnum <= end {
-                let matched = match matched.first() {
-                    Some(m) if m.line_number == lnum => {
-                        matched = &matched[1..];
-                        if let Some((s, e)) = m.range {
-                            Matched::Region(Region(s, e))
-                        } else {
-                            Matched::Yes
-                        }
+                let regions = match matched.split_first() {
+                    Some((m, ms)) if m.line_number == lnum => {
+                        matched = ms;
+                        Some(m.ranges.clone()) // XXX: Cannot move out ranges in line match
                     }
-                    _ => Matched::No,
+                    _ => None,
                 };
                 let line = String::from_utf8_lossy(bytes);
                 // Collect to `Vec` rather than handing HighlightIterator as-is. HighlightIterator takes ownership of Highlighter
                 // while the iteration. When the highlighter is stored in `self`, it means the iterator takes ownership of `self`.
-                self.draw_line(hl.highlight(line.as_ref()), lnum, matched)?;
+                self.draw_line(hl.highlight(line.as_ref()), lnum, regions)?;
 
                 if lnum == end {
+                    if self.first_only {
+                        break;
+                    }
                     if let Some(c) = chunks.next() {
                         self.draw_separator_line()?;
                         chunk = c;
@@ -768,9 +764,10 @@ impl<'file, W: Write> Drawer<'file, W> {
     }
 
     fn draw_header(&mut self, path: &Path) -> Result<()> {
-        self.draw_horizontal_line("─")?;
+        self.draw_horizontal_line(self.chars.horizontal)?;
         self.canvas.set_default_bg()?;
         let path = path.as_os_str().to_string_lossy();
+        self.canvas.set_default_fg()?;
         self.canvas.set_bold()?;
         write!(self.canvas, " {}", path)?;
         if self.background {
@@ -779,22 +776,21 @@ impl<'file, W: Write> Drawer<'file, W> {
         }
         self.canvas.draw_newline()?;
         if self.grid {
-            self.draw_horizontal_line("┬")?;
+            self.draw_horizontal_line(self.chars.down_and_horizontal)?;
         }
         Ok(())
     }
 
     fn draw_footer(&mut self) -> Result<()> {
         if self.grid {
-            self.draw_horizontal_line("┴")?;
+            self.draw_horizontal_line(self.chars.up_and_horizontal)?;
         }
         Ok(())
     }
 }
 
 fn load_themes(name: Option<&str>) -> Result<ThemeSet> {
-    let bat_defaults: ThemeSet =
-        bincode::deserialize_from(flate2::read::ZlibDecoder::new(THEME_SET_BIN))?;
+    let bat_defaults: ThemeSet = bincode::deserialize_from(ZlibDecoder::new(THEME_SET_BIN))?;
     match name {
         None => Ok(bat_defaults),
         Some(name) if bat_defaults.themes.contains_key(name) => Ok(bat_defaults),
@@ -810,6 +806,34 @@ fn load_themes(name: Option<&str>) -> Result<ThemeSet> {
     }
 }
 
+pub struct SyntectAssets {
+    pub syntax_set: SyntaxSet,
+    pub theme_set: ThemeSet,
+}
+
+impl SyntectAssets {
+    pub fn load(theme: Option<&str>) -> Result<Self> {
+        let syntax_set = bincode::deserialize_from(ZlibDecoder::new(SYNTAX_SET_BIN))?;
+        let theme_set = load_themes(theme)?;
+        Ok(Self {
+            syntax_set,
+            theme_set,
+        })
+    }
+}
+
+impl Clone for SyntectAssets {
+    fn clone(&self) -> Self {
+        let syntax_set = self.syntax_set.clone();
+        let mut theme_set = ThemeSet::new(); // ThemeSet does not implement Clone
+        theme_set.themes = self.theme_set.themes.clone();
+        Self {
+            syntax_set,
+            theme_set,
+        }
+    }
+}
+
 pub struct SyntectPrinter<'main, W>
 where
     for<'a> W: LockableWrite<'a>,
@@ -830,15 +854,24 @@ impl<'main, W> SyntectPrinter<'main, W>
 where
     for<'a> W: LockableWrite<'a>,
 {
-    pub fn new(out: W, opts: PrinterOptions<'main>) -> Result<Self> {
+    pub fn new(writer: W, opts: PrinterOptions<'main>) -> Result<Self> {
         Ok(Self {
-            writer: out,
-            syntaxes: bincode::deserialize_from(flate2::read::ZlibDecoder::new(SYNTAX_SET_BIN))?,
+            writer,
+            syntaxes: bincode::deserialize_from(ZlibDecoder::new(SYNTAX_SET_BIN))?,
             themes: load_themes(opts.theme)?,
             opts,
         })
     }
 
+    pub fn with_assets(assets: SyntectAssets, writer: W, opts: PrinterOptions<'main>) -> Self {
+        Self {
+            writer,
+            syntaxes: assets.syntax_set,
+            themes: assets.theme_set,
+            opts,
+        }
+    }
+
     pub fn writer_mut(&mut self) -> &mut W {
         &mut self.writer
     }
@@ -902,12 +935,17 @@ where
 mod tests {
     use super::*;
     use crate::chunk::{File, LineMatch};
+    use lazy_static::lazy_static;
     use std::cell::{RefCell, RefMut};
     use std::fmt;
     use std::fs;
     use std::mem;
     use std::path::PathBuf;
 
+    lazy_static! {
+        static ref ASSETS: SyntectAssets = SyntectAssets::load(None).unwrap();
+    }
+
     struct DummyStdoutLock<'a>(RefMut<'a, Vec<u8>>);
     impl<'a> Write for DummyStdoutLock<'a> {
         fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
@@ -943,11 +981,20 @@ mod tests {
                     let lnum = (idx + 1) as u64;
                     s = cmp::min(s, lnum);
                     e = cmp::max(e, lnum);
-                    if let (Some(start), Some(i)) = (line.find("*match to "), line.find(" line*")) {
+
+                    let mut ranges = vec![];
+                    let mut l = line;
+                    let mut base = 0;
+                    while let (Some(start), Some(i)) = (l.find("*match to "), l.find(" line*")) {
                         let end = i + " line*".len();
+                        ranges.push((base + start, base + end));
+                        l = &l[end..];
+                        base += end;
+                    }
+                    if ranges.len() > 0 {
                         lmats.push(LineMatch {
                             line_number: lnum,
-                            range: Some((start, end)),
+                            ranges,
                         });
                     }
                 } else {
@@ -1024,7 +1071,7 @@ mod tests {
             opts.term_width = 80;
             opts.color_support = TermColorSupport::True;
             f(&mut opts);
-            let mut printer = SyntectPrinter::new(stdout, opts).unwrap();
+            let mut printer = SyntectPrinter::with_assets(ASSETS.clone(), stdout, opts);
             printer.print(file).unwrap();
             let printed = mem::take(printer.writer_mut()).0.into_inner();
             let expected = read_expected_file(&expected_file);
@@ -1177,6 +1224,22 @@ mod tests {
             test_wrap_region_line_start(|_| {}),
             test_wrap_region_line_end(|_| {}),
             test_wrap_3_lines_emoji(|_| {}),
+            test_first_only(|o| {
+                o.first_only = true;
+            }),
+            test_ascii_lines_grid(|o| {
+                o.ascii_lines = true;
+            }),
+            test_ascii_lines_no_grid(|o| {
+                o.ascii_lines = true;
+                o.grid = false;
+            }),
+            test_multi_regions(|_| {}),
+            test_multi_regions_bg(|o| {
+                o.background_color = true;
+            }),
+            test_wrap_between_regions(|_| {}),
+            test_wrap_regions_japanese(|_| {}),
         );
     }
 
@@ -1220,7 +1283,7 @@ mod tests {
     fn test_error_write() {
         let file = sample_chunk("README.md");
         let opts = PrinterOptions::default();
-        let printer = SyntectPrinter::new(ErrorStdout, opts).unwrap();
+        let printer = SyntectPrinter::with_assets(ASSETS.clone(), ErrorStdout, opts);
         let err = printer.print(file).unwrap_err();
         assert_eq!(&format!("{}", err), "dummy error!", "message={}", err);
     }
@@ -1255,7 +1318,7 @@ mod tests {
         let file = File::new(PathBuf::from("x.txt"), vec![], vec![], vec![]);
         let opts = PrinterOptions::default();
         let stdout = DummyStdout(RefCell::new(vec![]));
-        let mut printer = SyntectPrinter::new(stdout, opts).unwrap();
+        let mut printer = SyntectPrinter::with_assets(ASSETS.clone(), stdout, opts);
         printer.print(file).unwrap();
         let printed = mem::take(printer.writer_mut()).0.into_inner();
         assert!(
@@ -1270,9 +1333,43 @@ mod tests {
         let file = sample_chunk("LICENSE.txt");
         let opts = PrinterOptions::default();
         let stdout = DummyStdout(RefCell::new(vec![]));
-        let mut printer = SyntectPrinter::new(stdout, opts).unwrap();
+        let mut printer = SyntectPrinter::with_assets(ASSETS.clone(), stdout, opts);
         printer.print(file).unwrap();
         let printed = mem::take(printer.writer_mut()).0.into_inner();
         assert!(!printed.is_empty());
     }
+
+    #[test]
+    fn test_adjacent_regions() {
+        let contents = b"this is test\n";
+        let ranges = (0..contents.len()).map(|i| (i, i + 1)).collect();
+        let lmats = vec![LineMatch {
+            line_number: 1,
+            ranges,
+        }];
+        let chunks = vec![(1, 1)];
+        let file = File::new(PathBuf::from("test.txt"), lmats, chunks, contents.to_vec());
+
+        let mut opts = PrinterOptions::default();
+        opts.color_support = TermColorSupport::True;
+        let stdout = DummyStdout(RefCell::new(vec![]));
+        let mut printer = SyntectPrinter::with_assets(ASSETS.clone(), stdout, opts);
+        printer.print(file).unwrap();
+
+        let printed = mem::take(printer.writer_mut()).0.into_inner();
+        let mut lines = printed.split_inclusive(|b| *b == b'\n');
+
+        // One region per one character, but color codes between adjacent regions are not inserted
+        let expected = b"\x1b[38;2;0;0;0m\x1b[48;2;255;231;146mthis is test";
+        let this_is_test_line = lines.nth(3).unwrap();
+        let found = this_is_test_line
+            .windows(expected.len())
+            .find(|s| s == expected)
+            .is_some();
+        assert!(
+            found,
+            "line={:?}",
+            std::str::from_utf8(this_is_test_line).unwrap()
+        );
+    }
 }
